from enum import IntEnum
from .packets import PacketIDs


class SysStatus(IntEnum):
    """系统状态码

    LOADING: 启动器正在加载
    LAUNCHING: 启动器正在启动
    RUNNING: 启动器正在运行
    NORMAL_EXIT: 正常退出
    FB_LAUNCH_EXC: FastBuilder 启动异常
    CRASHED_EXIT: 启动器崩溃退出
    NEED_RESTART: 需要重启
    """

    LOADING = 100
    LAUNCHING = 101
    RUNNING = 102
    NORMAL_EXIT = 103
    FB_LAUNCH_EXC = 104
    CRASHED_EXIT = 105
    NEED_RESTART = 106


TOOLDELTA_LOGO = """╔═════════════════════════════════════════════════════════════════════════╗
║§9████████╗ ██████╗  ██████╗ ██╗     §b██████╗ ███████╗██╗  ████████╗ █████╗ §r║
║§9╚══██╔══╝██╔═══██╗██╔═══██╗██║     §b██╔══██╗██╔════╝██║  ╚══██╔══╝██╔══██╗§r║
║§9   ██║   ██║   ██║██║   ██║██║     §b██║  ██║█████╗  ██║     ██║   ███████║§r║
║§9   ██║   ██║   ██║██║   ██║██║     §b██║  ██║██╔══╝  ██║     ██║   ██╔══██║§r║
║§9   ██║   ╚██████╔╝╚██████╔╝███████╗§b██████╔╝███████╗███████╗██║   ██║  ██║§r║
║§9   ╚═╝    ╚═════╝  ╚═════╝ ╚══════╝§b╚═════╝ ╚══════╝╚══════╝╚═╝   ╚═╝  ╚═╝§r║
╚═════════════════════════════════════════════════════════════════════════╝§r"""
"ToolDelta标志"

TOOLDELTA_LOGO_witout_colors = """╔═════════════════════════════════════════════════════════════════════════╗
║████████╗ ██████╗  ██████╗ ██╗     ██████╗ ███████╗██╗  ████████╗ █████╗ ║
║╚══██╔══╝██╔═══██╗██╔═══██╗██║     ██╔══██╗██╔════╝██║  ╚══██╔══╝██╔══██╗║
║   ██║   ██║   ██║██║   ██║██║     ██║  ██║█████╗  ██║     ██║   ███████║║
║   ██║   ██║   ██║██║   ██║██║     ██║  ██║██╔══╝  ██║     ██║   ██╔══██║║
║   ██║   ╚██████╔╝╚██████╔╝███████╗██████╔╝███████╗███████╗██║   ██║  ██║║
║   ╚═╝    ╚═════╝  ╚═════╝ ╚══════╝╚═════╝ ╚══════╝╚══════╝╚═╝   ╚═╝  ╚═╝║
╚═════════════════════════════════════════════════════════════════════════╝"""
"无颜色的ToolDelta标志"
TOOLDELTA_LOGO_mode = [
    [0, TOOLDELTA_LOGO],
    [1, TOOLDELTA_LOGO_witout_colors, (0, 100, 255), (138, 43, 226)],
    [1, TOOLDELTA_LOGO_witout_colors, (34, 139, 34), (144, 238, 144)],
    [1, TOOLDELTA_LOGO_witout_colors, (255, 165, 0), (128, 0, 128)],
    [1, TOOLDELTA_LOGO_witout_colors, (255, 182, 193), (221, 160, 221)],
]
FBLIKE_APIS = [
    "%s/api/phoenix/login",
    "%s/api/new",
    "%s",
]
"验证服务器: FastBuilder API 列表"

AUTH_SERVERS = [
    ("§7FastBuilder 官方验证服务器 §c✘不再可用", "https://user.fastbuilder.pro"),
    ("§7咕咕酱 FB验证服务器 §c✘不再可用", "https://liliya233.uk"),
    ("NetHard 验证服务器 §a✔可用", "https://nv1.nethard.pro")
]
"验证服务器列表"

TOOLDELTA_PLUGIN_DIR = "插件文件"
"插件文件路径"

TOOLDELTA_CLASSIC_PLUGIN = "ToolDelta类式插件"
"插件文件: ToolDelta 类式插件 路径"

TOOLDELTA_PLUGIN_CFG_DIR = "插件配置文件"
"插件配置文件文件夹路径"

TOOLDELTA_PLUGIN_DATA_DIR = "插件数据文件"
"插件数据文件文件夹路径"

PLUGIN_TYPE_MAPPING = {
    "classic": TOOLDELTA_CLASSIC_PLUGIN,
}
"插件属性名映射"

TDSPECIFIC_MIRROR = "https://github.tooldelta.top"
"ToolDelta的github镜像"

TDSPECIFIC_GITHUB_DOWNLOAD_URL = (
    TDSPECIFIC_MIRROR + "/https://raw.githubusercontent.com"
)
"ToolDelta镜像: github文件下载"

TDREPO_URL = (
    f"{TDSPECIFIC_MIRROR}/https://api.github.com/repos/ToolDelta-Basic/ToolDelta"
)
"ToolDelta镜像: github项目仓库"

TDDEPENDENCY_REPO_RAW = "ToolDelta-Basic/DependencyLibrary"
TDDEPENDENCY_REPO = "https://raw.githubusercontent.com/" + TDDEPENDENCY_REPO_RAW

PLUGIN_MARKET_SOURCE_OFFICIAL = "https://github.tooldelta.top/https://raw.githubusercontent.com/ToolDelta-Basic/PluginMarket/main"


class LaunchMode(IntEnum):
    LAUNCH_MODE_NEOMG_ACCESS_POINT = 1
    LAUNCH_MODE_NEOMG_ACCESS_POINT_REMOTE = 2
    LAUNCH_MODE_NEOMG_LAUNCH = 3
    LAUNCH_MODE_EULOGIST = 4


INTERNAL_LISTEN_PACKETS: set[PacketIDs] = {
    PacketIDs.Text,
    PacketIDs.PlayerList,
    PacketIDs.CommandOutput,
    PacketIDs.UpdateAbilities,
}



import json
import os
from typing import Any
from . import fmts

PathLike = str | os.PathLike[str]

def safe_write(filepath: PathLike, obj: Any, indent=2):
    """
    安全地写入文件, 防止文件在写入途中被强行中止。

    Args:
        filepath (TextIOWrapper): 文件路径
        obj (Any): 写入的 json 待序列化对象
    """
    retry_times = 0
    while 1:
        try:
            bak_name = str(filepath) + ".bak"
            content = json.dumps(obj, indent=indent, ensure_ascii=False)
            # 1. make backup
            with open(bak_name, "w", encoding="utf-8") as fp:
                fp.write(content)
            # Seemed unneccessary
            # # 2. write orig file
            # with open(filepath, "w", encoding="utf-8") as fp:
            #     fp.write(content)
            break
        except BaseException as err:
            # 防止写入过程中被退出, 导致文件只写了一半
            retry_times += 1
            if retry_times >= 50:
                fmts.print_err(f"文件第 50 次写入失败: {err}")
                raise
            fmts.print_war(f"文件在写入时遭到强行中止 ({err}), 重试第 {retry_times} 次")
            continue
    try:
        os.remove(filepath)
    except FileNotFoundError:
        pass
    os.rename(bak_name, filepath)



from dataclasses import dataclass


@dataclass
class QueriedInventoryItemEnchant:
    level: int
    name: str
    type: int

    @classmethod
    def from_dict(cls, dic: dict):
        return cls(level=dic["level"], name=dic["name"], type=dic["type"])


@dataclass
class QueriedInventoryItem:
    aux: int
    enchantments: list[QueriedInventoryItemEnchant]
    freeStackSize: int
    id: str
    maxStackSize: int
    namespace: str
    stackSize: int

    @classmethod
    def from_dict(cls, dic: dict):
        return cls(
            aux=dic["aux"],
            enchantments=[
                QueriedInventoryItemEnchant.from_dict(e) for e in dic["enchantments"]
            ],
            freeStackSize=dic["freeStackSize"],
            id=dic["id"],
            maxStackSize=dic["maxStackSize"],
            namespace=dic["namespace"],
            stackSize=dic["stackSize"],
        )


@dataclass
class QueriedInventory:
    first: int
    last: int
    slotCount: int
    slots: list[QueriedInventoryItem | None]

    @classmethod
    def from_dict(cls, dic: dict):
        return cls(
            first=dic["first"],
            last=dic["last"],
            slotCount=dic["slotCount"],
            slots=[
                QueriedInventoryItem.from_dict(slot) if slot else None
                for slot in dic["slots"]
            ],
        )



import os
import sys
from tooldelta import constants

def init_dirs():
        """初始化系统文件夹"""
        os.makedirs(
            os.path.join("插件文件", constants.TOOLDELTA_CLASSIC_PLUGIN), exist_ok=True
        )
        os.makedirs("插件配置文件", exist_ok=True)
        os.makedirs(os.path.join("tooldelta", "bin"), exist_ok=True)
        os.makedirs(os.path.join("插件数据文件", "game_texts"), exist_ok=True)
        if sys.platform == "win32":
            win_create_batch_file()


def win_create_batch_file():
    if not os.path.isfile("点我启动.bat"):
        argv = sys.argv.copy()
        if argv[0].endswith(".py"):
            argv.insert(0, "python")
        exec_cmd = " ".join(argv)
        with open("点我启动.bat", "w") as f:
            f.write(f"@echo off\n{exec_cmd}\npause")



import grpc

from ...constants import tooldelta_cfg
from ...utils.cfg import get_cfg
from ...constants import SysStatus
from ...utils import fmts, sys_args
from .fateark_libs import core_conn as fateark_core
from .fateark_access_point import FrameFateArk


class FrameFateArkIndirect(FrameFateArk):
    launch_type = "FateArk"
    proc = None

    def __init__(self) -> None:
        super().__init__()


    def launch(self):
        cfgs = get_cfg("ToolDelta基本配置.json", tooldelta_cfg.LAUNCH_CFG_STD)
        openat_addr = (
            sys_args.sys_args_to_dict().get("access-point-port")
            or cfgs.get("FateArk远程接入点模式", {}).get(
                "远程连接端口", "tcp://127.0.0.1:24020"
            )
        ).removeprefix("tcp://")
        self.update_status(SysStatus.LAUNCHING)
        fmts.print_suc(f"将在 {openat_addr} 端口启动 FateArk 接入点")
        fateark_core.connect(openat_addr)
        self._message_show_thread()
        try:
            status, _, err_msg = fateark_core.login(
                self.auth_server,
                self.fbToken,
                str(self.serverNumber),
                self.serverPassword,
            )
        except grpc.RpcError as err:
            self.update_status(SysStatus.CRASHED_EXIT)
            return SystemError(f"FateArk 与 ToolDelta 断开连接: {err.details()}")
        if status != 0:
            self.update_status(SysStatus.CRASHED_EXIT)
            return SystemError(f"FateArk 登录失败: {err_msg}")
        fmts.print_suc("FateArk 已连接")
        self.update_status(SysStatus.RUNNING)
        fateark_core.set_listen_packets(set(self.need_listen_packets))
        self._packets_handler_thread()
        self._exec_launched_listen_cbs()
        self.wait_crashed()
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出")
        if self.status == SysStatus.CRASHED_EXIT:
            return Exception("接入点进程已崩溃")
        return SystemError("未知的退出状态")
